package org.lucci.madhoc.broadcast;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.lucci.madhoc.messaging.Message;
import org.lucci.madhoc.messaging.MessageInformation;
import org.lucci.madhoc.messaging.TransferableObject;
import org.lucci.madhoc.network.Application;
import org.lucci.madhoc.network.Station;
import org.lucci.madhoc.simulation.Configurable;


/*
 * 
 * @author luc.hogie
 * Created on Jun 20, 2004
 * 
 * This class implements the basic behavior of a broadcasting application: 
 * 
 */


public abstract class BroadcastingProtocol extends Application implements Configurable
{
	protected Map<TransferableObject, LocalInfo> hostedObjects = new HashMap<TransferableObject, LocalInfo>();
	private int numberOfReactiveEmissions = 0;
	private int numberOfProactiveEmissions = 0;


	/**
	 * Return if the emission of the given ad is worth or not.
	 * By default, it is assumed to be worth if the RAD has expired.
	 * Broadcasting protocols may redefine this rule by redefine this method. 
	 * @param ad
	 * @return
	 */
	protected abstract boolean isBroadcastWorth(TransferableObject object);

	protected void messageJustReceived(Message msg)
	{
		LocalInfo info = getLocalInfo(msg.getContent());
		BroadcastInformation broadcastInfo = info.broadcastInformation;

		// if the message has been received in the past, the new number of host
		// should not be considered, as we consider the shortest one
		if (info.redundantReceptionCount == 0)
		{
			broadcastInfo.addHops();
		}
		
		if (broadcastInfo.getHops() >= broadcastInfo.getMaxHop())
		{
			info.decision = 0;
		}
	}
	
	protected abstract void messageJustBeingSent(Message msg);
	protected abstract boolean isMultipleRebroadcastAllowed();

	
	public LocalInfo getLocalInfo(Object object)
	{
		return (LocalInfo) hostedObjects.get(object);
	}

	public Map<TransferableObject, LocalInfo> getHostedObjectsMap()
	{
		return java.util.Collections.unmodifiableMap(hostedObjects);
	}

	public int getMemoryRequired()
	{
		int sz = 0;
		
		for (TransferableObject o : getHostedObjectsMap().keySet())
		{
			LocalInfo info = getLocalInfo(o);
			sz += o.getSizeInBytes() + info.getSizeInBytes();
		}
		
		return sz;
	}


	public void doIt(double time)
	{
//		BroadcastingSimulationApplication app = (BroadcastingSimulationApplication) getSimulationApplication();
//		
//		if (app.getAverageCoverageRate() > 0.8)
//		{
//			Iterator adIt = getBroadcastMessagesMap().keySet().iterator();
//			
//			while (adIt.hasNext())
//			{
//				BroadcastMessage m = (BroadcastMessage) adIt.next();
//				m.setExpirationDate(getSimulatedTimeMillis());
//			}
//		}
		
		for (Message msg : getIncomingMessages())
		{
            if (msg.getType() == Message.APPLICATION_MESSAGE)
            {
                TransferableObject object = msg.getContent();
                LocalInfo info = getLocalInfo(object);

                // if the ad is received for the first time
                if (info == null)
                {
                    info = instantiateLocalInfo();
                    info.object = object;
                    info.message = msg;
                    info.broadcastInformation = (BroadcastInformation) msg.getMessageInformation();
                    hostedObjects.put(object, info);
                }
                else
                {
                    ++info.redundantReceptionCount;
                }

                messageJustReceived(msg);
            }
		}

//		flushOutdatedAds();
		deliverMessagesToLocalApplications();
		broadcastAdsWhenWorth();
	}


	private void deliverMessagesToLocalApplications()
	{
		for (TransferableObject object : getHostedObjectsMap().keySet())
		{
			LocalInfo info = getLocalInfo(object);
			
			// if the object didn't arrive by message (but was created created here, for example)
			if (info.message != null)
			{
				BroadcastInformation broadcastInfo = info.broadcastInformation;
				Application targetApplication = (Application) getComputer().getApplicationMap().getValue(broadcastInfo.getTargetApplicationClass());
				
				if (targetApplication != null)
				{
					try
					{
						Message msg = (Message) info.message.clone();
						msg.getRecipientApplications().add(targetApplication);
						getComputer().getNetworkingUnit().getIncomingMessageQueue().add(msg);
					}
					catch (IOException ex)
					{
						// the buffer is full, we'll try later
					}
				}
			}
		}
	}


	private void broadcastAdsWhenWorth()
	{
		// if the device has some connections, it may try to broadcast its ads
		Iterator objects = getHostedObjectsMap().keySet().iterator();

		while ( objects.hasNext() )
		{
			TransferableObject object = (TransferableObject) objects.next();
			LocalInfo info = getLocalInfo(object);

			// if the message has already been sent and if multiple broadcasting is not allowed
			// don't even ask to protocol to decide if an emission would be worth or not
			if ((info.emissionCount == 0 || isMultipleRebroadcastAllowed()))
			{
				if (isBroadcastWorth(object))
				{
					Message msg = new Message();
					msg.setContent(object);
					
					// this is a broadcasting, all the nodes around are targeted
					for (Station node : getComputer().getNetworkingUnit().getNeighborhood())
					{
						Application app = (Application) node.getApplicationMap().getValue(getClass());

						if (app != null)
						{
							msg.getRecipientApplications().add(app);
						}
					}

					if (info.broadcastInformation == null)
					{
						msg.setMessageInformation(instantiateBroadcastInformation());
					}
					else
					{
						msg.setMessageInformation((MessageInformation) info.broadcastInformation.clone());
					}

					msg.setSourceStationApplication(this);

					try
					{
						getComputer().getNetworkingUnit().getOutgoingMessageQueue().add(msg);
						++info.emissionCount;

						if (info.sendStrategy == LocalInfo.NOT_SENT)
						{
							info.sendStrategy = LocalInfo.REACTIVE;
						}

						if (info.sendStrategy  == LocalInfo.REACTIVE)
						{
							++numberOfReactiveEmissions;							
						}
						else if (info.sendStrategy  == LocalInfo.PROACTIVE)
						{
							++numberOfProactiveEmissions;							
						}
						else
						{
							throw new IllegalStateException();
						}

						messageJustBeingSent(msg);
						info.sendStrategy = LocalInfo.NOT_SENT;
					}
					catch (IOException e)
					{
						e.printStackTrace();
					}
				}
			}
		}
	}






	/**
	 * Subclasses may redefine this method if they want to manage more information
	 * about the ad they host.
	 */
	protected LocalInfo instantiateLocalInfo()
	{
		return new LocalInfo();
	}

	protected BroadcastInformation instantiateBroadcastInformation()
	{
		return new BroadcastInformation();
	}

	public void addObject(TransferableObject object)
	{
		LocalInfo info = instantiateLocalInfo();
		info.retransmissionDate = getSimulatedTime();
		info.broadcastInformation = instantiateBroadcastInformation();
		BroadcastingMonitor monitor = (BroadcastingMonitor) getMonitor();
		info.broadcastInformation.setMaxHop(monitor.getBroadcastMaxHop());
		hostedObjects.put(object, info);
	}


//	private void flushOutdatedAds()
//	{
//		Iterator msgs = hostedObjects.keySet().iterator();
//		
//		while ( msgs.hasNext() )
//		{
//			Object object = msgs.next();
//			BroadcastInformation msg = (BroadcastInformation) null; 
//			
//			if ( msg.getExpirationDate() < getSimulatedTimeMillis())
//			{
//				msgs.remove();
//			}
//		}
//	}


	public double getAverageNumberOfEmissions()
	{
		double sum = 0;
		Iterator adInfoIterator = getHostedObjectsMap().values().iterator();
		
		while (adInfoIterator.hasNext())
		{
			LocalInfo info = (LocalInfo) adInfoIterator.next();
			sum += info.emissionCount;
		}
		
		return sum / getHostedObjectsMap().keySet().size();
	}

	public int getNumberOfProactiveEmissions()
	{
		return numberOfProactiveEmissions;
	}
	public int getNumberOfReactiveEmissions()
	{
		return numberOfReactiveEmissions;
	}
}
