package org.lucci.madhoc.broadcast.batch;


import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.lucci.config.TypedConfiguration;
import org.lucci.madhoc.broadcast.BroadcastingMonitor;
import org.lucci.madhoc.broadcast.BroadcastingProtocol;
import org.lucci.madhoc.broadcast.RandomDelayBasedBroadcastingProtocol;
import org.lucci.madhoc.broadcast.impl.research.dfcn.DFCN;
import org.lucci.madhoc.broadcast.impl.standard.MultiPointRelaying;
import org.lucci.madhoc.broadcast.impl.standard.AHBP_EX;
import org.lucci.madhoc.broadcast.measure.AverageNumberOfEmissionMeasure;
import org.lucci.madhoc.config.ConfigurationKeys;
import org.lucci.madhoc.script.Batch;
import org.lucci.madhoc.simulation.MadhocSimulation;
import org.lucci.madhoc.simulation.measure.MeasureHistory;
import org.lucci.madhoc.simulation.measure.NumericalMeasureHistory;
import org.lucci.madhoc.simulation.measure.NumericalSensor;
import org.lucci.madhoc.simulation.measure.Sensor;
import org.lucci.madhoc.simulation.projection.IdentityProjection;
import org.lucci.math.Utilities;
import org.lucci.math.relation.DefaultRelation;
import org.lucci.math.relation.Function;
import org.lucci.math.relation.Relation;
import org.lucci.util.Collections;

/*
 * Created on Aug 18, 2004
 */

/**
 * 
 * For each broadcasting protocol:
 * vary the density of mobile stations in [500, 10000]
 * and measures the hosting rate, emission efficiency
 * each mesure being the average of 10 measures
 * 
 * 
 * 
 * @author luc.hogie
 */
public class CompareDFCNToTheOthers extends Batch
{

	public static void main(String[] args)
	    throws Throwable
	{
		new CompareDFCNToTheOthers();
	}

	public CompareDFCNToTheOthers()
	    throws Throwable
	{
		test();
	}

	private void test()
	    throws Throwable
	{
		println("Starting tests...");
		testProtocol(org.lucci.madhoc.broadcast.impl.standard.SimpleFlooding.class);
		testProtocol(org.lucci.madhoc.broadcast.impl.standard.FloodingWithSelfPruning.class);
		testProtocol(org.lucci.madhoc.broadcast.impl.standard.SBA.class);
        testProtocol(MultiPointRelaying.class);
        testProtocol(AHBP_EX.class);
		testProtocol(DFCN.class);
//		testProtocol(ProbabilisticScheme.class);
//
//		testProtocol(SDFCN.SDFCN_00.class);
//		testProtocol(SDFCN.SDFCN_02.class);
//		testProtocol(SDFCN.SDFCN_04.class);
//		testProtocol(SDFCN.SDFCN_06.class);
//		testProtocol(SDFCN.SDFCN_08.class);
//		testProtocol(SDFCN.SDFCN_10.class);
//		testProtocol(Proba.class);
//		testProtocol(FreeDFCN.class);
//		testProtocol(PurelyProactiveDFCN.class);
//		testProtocol(PurelyReactiveDFCN.class);
//		testProtocol(FreeProactiveDFCN.class);
		println("Completed.");

        List<Parameter> parameters = new Vector<Parameter>(); 

        parameters.add(new Parameter("network_environment", new String[] {
                "org.lucci.madhoc.network.env.open_area.OpenAreaEnvironment",
                "org.lucci.madhoc.network.env.mall.HumanEnvironment"
            }));

        parameters.add(new Parameter("broadcasting_protocol_class", new String[] {
                "org.lucci.madhoc.broadcast.impl.standard.FloodingWithSelfPruning",
                "org.lucci.madhoc.broadcast.impl.standard.SimpleFlooding"
            }));
	}


	/**
	 * For the given protocol, returns the evolution of each measure with the variation of the density.
	 * @param protocol
	 * @return
	 */
	private void testProtocol(Class protocol)
        throws Throwable
    {
        int numberOfRuns = 1;
        int[] steps = {100, 500, 1000, 2000};

        String protocolName = protocol.getName().substring(protocol.getName().lastIndexOf(".") + 1 );
		println("\ttesting protocol " + protocolName);

        
		Map<Sensor, Function> sensor_function = new HashMap<Sensor, Function>(); 

		for (int i = 0; i < steps.length; ++i)
		{
			int stationDensity = steps[i];
            Map<Sensor, AvgAndStandDev> measure_value = nRuns(numberOfRuns, protocol, stationDensity, new int[] {3, 5});
			
            if (measure_value != null)
            {
    			for (Sensor sensor : measure_value.keySet())
    			{
    				Function function = (Function) sensor_function.get(sensor);
    				
    				if (function == null)
    				{
    					function = new Function();
    					function.setName(sensor.getName());
                        sensor_function.put(sensor, function);
    				}
    				
    				function.add(stationDensity, measure_value.get(sensor));
    			}
            }
		}

		for (Sensor sensor : sensor_function.keySet())
		{
			Function function = (Function) sensor_function.get(sensor);
			File file = getDataFile(new String[] {protocolName, sensor.getName()}, new File("."));
			println("\t\twriting file " + file.getAbsolutePath());
            
            if (!file.getParentFile().exists())
            {
                file.getParentFile().mkdirs();
            }
            
			saveFunctionToFile(function, file);
		}
	}
			
	
	//Map<Measure><Value>
	private Map<Sensor, AvgAndStandDev> nRuns(int numberOfRuns, Class protocol, int density, int[] velocityInterval)
            throws Throwable
	{
		println("\t\twith density=" + density);
		Relation sensor_measures = new DefaultRelation(); 

		for (int i = 1; i <= numberOfRuns; ++i)
		{
			println("\t\t\trun n" + i);
            Map<Sensor, Double> usableMeasures = oneRun(protocol, density, velocityInterval);

            if (usableMeasures == null)
            {
                return null;
            }

            for (Sensor sensor : usableMeasures.keySet())
			{
				sensor_measures.add(sensor, usableMeasures.get(sensor));
			}
		}

		println("\t\t\tProcessing average");
		Map<Sensor, AvgAndStandDev> sensor_avgMeasure = new HashMap<Sensor, AvgAndStandDev>();
		
		for (Sensor sensor : (Collection<Sensor>) sensor_measures.getKeys())
		{
			Collection measures = sensor_measures.getValues(sensor);
			double average = Utilities.getAverage(measures);
			double standardDeviation = Utilities.getStandardDeviation(measures);
            sensor_avgMeasure.put(sensor, new AvgAndStandDev(average, standardDeviation));
		}
		
		return sensor_avgMeasure;
	}

	
	private Map<Sensor, Double> oneRun(Class protocolClass, int mobileStationDensity, int[] velocityInterval)
        throws Throwable
	{
        println("\t\t\t\tInitialization of the simulation");
        ConfigurationKeys configurationKeys = new ConfigurationKeys();
        configurationKeys.simulation_resolution = "1";
        configurationKeys.broadcast_rad_interval = "0, 5";
        configurationKeys.random_waypoint_mobility_velocity_interval = velocityInterval[0] + ", " + velocityInterval[1];
        configurationKeys.network_phone_density = String.valueOf(mobileStationDensity);
        configurationKeys.broadcasting_protocol_class = protocolClass.getName();
        configurationKeys.network_environment="org.lucci.madhoc.network.env.mall.HumanEnvironment";

        
        TypedConfiguration configuration = new TypedConfiguration();
        configuration.load(configurationKeys);

        MadhocSimulation simulation = new MadhocSimulation();
        simulation.setConfiguration(configuration);
        simulation.configure();

        BroadcastingMonitor broadcastingMonitor = (BroadcastingMonitor) simulation.getNetwork().getMonitorMap().get(BroadcastingMonitor.class);
        RandomDelayBasedBroadcastingProtocol protocol = (RandomDelayBasedBroadcastingProtocol) Collections.getRandomObject(broadcastingMonitor.getDeployedApplications(), simulation.getRandomNumberGenerator().getRandom());
        NumericalMeasureHistory measureHistory = (NumericalMeasureHistory) simulation.getNetwork().getProjectionMap().get(IdentityProjection.class).getMeasureHistoryMap().get(AverageNumberOfEmissionMeasure.class);

        while (broadcastingMonitor.getAverageCoverageRate(simulation.getNetwork().getStations()) < 0.8)
		{
            // if the broadcasting process do not operate anymore, there is no need in continuing the simulation
            if (simulation.getSimulatedTime() > protocol.getMaximumDelay())
            {
                if (!broadcastingMonitor.isMultipleRebroadcastingAllowed()
                        && measureHistory.getVariationInTheLastSeconds(protocol
                                .getMaximumDelay()) == 0)
                {
                    return null;
                }
            }
            
            if (simulation.getSimulatedTime() > 60)
			{
				return null;
			}

            println("\t\t\t\tstarting iteration " + simulation.getIteration());
			simulation.iterate();
		}
		
        // select and returns the sensor that can be used
        Map<Sensor, Double> usableMeasures = new HashMap<Sensor, Double>();
        
        for(Sensor sensor : simulation.findAllSensors())
        {
            if (sensor instanceof NumericalSensor)
            {
                MeasureHistory history = simulation.getNetwork().getProjectionMap().get(IdentityProjection.class).getMeasureHistoryMap().get(sensor.getClass());
                
                if (!history.getValues().isEmpty())
                {
                    usableMeasures.put(sensor, (Double) history.getLastValue());
                }
            }
        }
        
        return usableMeasures;
	}
}
