package org.lucci.madhoc.broadcast.batch;

import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.lucci.madhoc.broadcast.BroadcastingMonitor;
import org.lucci.madhoc.env.mall.MallEnvironment;
import org.lucci.madhoc.network.Network;
import org.lucci.madhoc.network.monitor.NetworkMonitor;
import org.lucci.madhoc.network.util.Graph;
import org.lucci.madhoc.script.Batch;
import org.lucci.madhoc.simulation.MadhocSimulation;
import org.lucci.madhoc.simulation.measure.Sensor;
import org.lucci.madhoc.simulation.measure.NumericalSensor;
import org.lucci.math.Utilities;
import org.lucci.math.relation.DefaultRelation;
import org.lucci.math.relation.Function;
import org.lucci.math.relation.Relation;

/*
 * Created on Aug 18, 2004
 */

/**
 * 
 * For each broadcasting protocol:
 * vary the density of mobile stations in [500, 10000]
 * and measures the hosting rate, emission efficiency
 * each mesure being the average of 10 measures
 * 
 * 
 * 
 * @author luc.hogie
 */
public class EvaluateLightweightBroadcasting extends Batch
{
	public static void main(String[] args)
	{
		new EvaluateLightweightBroadcasting();
	}
	
	public EvaluateLightweightBroadcasting()
	{
		test();
	}

	private void test()
	{
		println("Starting tests...");
 		Iterator protocols = new BroadcastingMonitor().getAvailableBroadcastingProtocols().iterator();
		
		while (protocols.hasNext())
		{
			Class protocolClass = (Class) protocols.next();
			testProtocol(protocolClass);
		}
		
		println("Completed.");
	}
	

	/**
	 * For the given protocol, returns the evolution of each measure with the variation of the density.
	 * @param protocol
	 * @return
	 */
//	private Map<Measure, Function> testProtocol(Class protocol)
	private void testProtocol(Class protocol)
	{
		String protocolName = protocol.getName().substring(protocol.getName().lastIndexOf(".") + 1 );
		println("\ttesting protocol " + protocolName);

		Relation measure_function = new DefaultRelation(); 
		int step = 1000;
		
		for (int stationDensity = 100; stationDensity <= 5000; stationDensity += step)
		{
			Map measure_value = testProtocol(protocol, stationDensity, 3);
			
			if (measure_value == null)
			{
			}
			else
			{
				Iterator measureIterator = measure_value.keySet().iterator();
				
				while (measureIterator.hasNext())
				{
					Sensor measure = (Sensor) measureIterator.next();
					Function function = (Function) measure_function.getValue(measure);
					
					if (function == null)
					{
						function = new Function();
						measure_function.add(measure, function);
					}
					
					function.add(new Double(stationDensity), measure_value.get(measure));
				}
			}
		}

		Iterator measureIterator = measure_function.getKeys().iterator();
		
		while (measureIterator.hasNext())
		{
			Sensor measure = (Sensor) measureIterator.next();
			Function function = (Function) measure_function.getValue(measure);
			File file = getDataFile(new String[] {protocolName, measure.getName()}, new File("."));
			println("\t\twriting file " + file.getAbsolutePath());
			saveFunctionToFile(function, file);
		}
}
			
	
	//Map<Measure><Double>
	private Map testProtocol(Class protocol, int density, int simulationCount)
	{
		println("\t\twith density=" + density);
		Relation measure_values = new DefaultRelation(); 

		for (int i = 0; i < simulationCount; ++i)
		{
			println("\t\t\ttest " + i);
			Collection relevantMeasures = testProtocolOnce(protocol, density);
			
			if (relevantMeasures == null)
			{
				return null;
			}
			else
			{
				Iterator measureIterator = relevantMeasures.iterator();
				
				while (measureIterator.hasNext())
				{
					Sensor measure = (Sensor) measureIterator.next();
					measure_values.add(measure, measure.getLastValue());
				}
			}
		}

		println("\t\t\tProcessing average");
		Map measure_value = new HashMap();
		Iterator measureIterator = measure_values.getKeys().iterator();
		
		while (measureIterator.hasNext())
		{
			Sensor measure = (Sensor) measureIterator.next();
			Collection values = measure_values.getValues(measure);
			double average = Utilities.getAverage(values);
			measure_value.put(measure, new Double(average));
		}
		
		return measure_value;
	}

	private Collection testProtocolOnce(Class protocol, int density)
	{
		MadhocSimulation simulation = carryOutSimulation(protocol, density);
		
		if (simulation == null)
		{
			return null;
		}
		else
		{
			Collection relevantMeasures = new Vector();
			Iterator measureIterator = simulation.findAllSensors().iterator();
			
			while (measureIterator.hasNext())
			{
				Sensor measure = (Sensor) measureIterator.next();
				
				if (measure instanceof NumericalSensor)
				{
					if (measure.getLastValue() != null)
					{
						relevantMeasures.add(measure);
					}
				}
			}
			
			return relevantMeasures;
		}
	}
	
	private MadhocSimulation carryOutSimulation(Class protocol, int mobileStationDensity)
	{
		// creates the network, which is initially empty
		Network net = new Network();
		
		// define a natural environment for the network
		// this environment will have a great impact on the structure of the network
		// for instance, the city environment will incorporate walls that will prevent the
		// radio waves from propagating
		// here we define a 1 square kilometer (1 million square meter)
		// open area (which could be a field) 
//		net.setEnvironment(new OpenAreaWithHotspotsEnvironment(1000000, 100));
		net.setNetworkEnvironment(new MallEnvironment(250000, 100));

		// creates the stations that will populate the network
		net.createStations(mobileStationDensity);
		
		if (!Graph.findUnreachableNetworkTypes(net).isEmpty())
			System.out.println("UNREACHABLE");
		
		// the stations have been located randomly on the environment
		// it might be interested to let them run for a while in order
		// to stabilize the network graph
//		net.initializeStationLocation();

		// once we consider the location of the station has been initialized
		// we establish the connections that define the network itself
		net.createNewConnections();

		// the network will be handled by a simulation
		MadhocSimulation simulation = new MadhocSimulation();
		simulation.setName("test protocol" + protocol.getName());
		simulation.setNetwork(net);

		// the broadcasting application is what we want to run and observe


		// it is now time to define the applications that will run on the network
		{
			Collection applications = new Vector();

			{
				// the "core" application is a basic application that aim at extracting some
				// general purpose information on the simulation process
				applications.add(new NetworkMonitor());
			}
			{
				// define the broadcasting strategy we want to carry out
				BroadcastingMonitor broadcastingApp = new BroadcastingMonitor();
				broadcastingApp.setStationApplicationClass(protocol);
				broadcastingApp.setMode(BroadcastingMonitor.ONE_TO_ALL);
				applications.add(broadcastingApp);
			}

			// the simulation knows the application is has to carry out
			// now let deploy them over the network
			simulation.deployApplications(applications);
		}

		// at this step the simulation is ready to be used
		// the user can get information on the simulation process via the Simulation.getMeasures() method,
		// via the methods provided by the various objects (Simulation, Network, Station...)
		// or via the discrete event file, depending on his own needs
		BroadcastingMonitor broadcastingApp = (BroadcastingMonitor) simulation.getMonitorMap().get(BroadcastingMonitor.class);

		
		while (simulation.getSimulatedTime() < 3600000)
		{
			simulation.iterate();
		}
		
		return simulation;
	}
}
