package org.lucci.madhoc.broadcast.batch;


import java.lang.reflect.Constructor;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

import org.lucci.config.Configuration;
import org.lucci.madhoc.broadcast.BroadcastingMonitor;
import org.lucci.madhoc.broadcast.measure.AverageNumberOfEmissionMeasure;
import org.lucci.madhoc.env.NetworkEnvironment;
import org.lucci.madhoc.network.Network;
import org.lucci.madhoc.network.Application;
import org.lucci.madhoc.network.monitor.NetworkMonitor;
import org.lucci.madhoc.script.Batch;
import org.lucci.madhoc.simulation.MadhocSimulation;
import org.lucci.madhoc.simulation.Monitor;
import org.lucci.madhoc.simulation.measure.Sensor;
import org.lucci.madhoc.simulation.measure.MeasureUtility;
import org.lucci.madhoc.simulation.measure.NumericalSensor;
import org.lucci.madhoc.simulation.monitor.SimulationMonitor;

/*
 * Created on Aug 18, 2004
 */

/**
 * 
 * For each broadcasting protocol:
 * vary the density of mobile stations in [500, 10000] 
 * and measures the hosting rate, emission efficiency
 * each mesure being the average of 10 measures
 * 
 * 
 * 
 * @author luc.hogie
 */
public class HundredPCCoverage extends Batch
{
	public static void main(String[] args)
		throws Throwable
	{
		if (args.length == 0)
		{
			Configuration config = new Configuration();
			config.load(System.in);

			try
			{
				new HundredPCCoverage(config);
			}
			catch (ClassNotFoundException ex)
			{
				ex.printStackTrace();
				System.err.println("Did you add 'toools.jar', 'madhoc.jar' and 'up.jar' in the classpath?");
			}
			catch (NoClassDefFoundError ex)
			{
				ex.printStackTrace();
				System.err.println("Did you add 'toools.jar', 'madhoc.jar' and 'up.jar' in the classpath?");
			}
		}
		else
		{
			System.err.println("No arguments allowed. The config should be piped to standard input. Example: cat config.txt | java org.lucci.madhoc_broadcast.batch");
		}
	}
	
	public  HundredPCCoverage(Configuration config)
		throws Throwable
	{
			// creates the network, which is initially empty
		Network net = new Network();
		
		// define a natural environment for the network
		// this environment will have a great impact on the structure of the network
		// for instance, the city environment will incorporate walls that will prevent the
		// radio waves from propagating
		// here we define a 1 square kilometer (1 million square meter)
		// open area (which could be a field)
		Class environmentClass = Class.forName((String) config.getValue("environment"));
		Constructor constructor = environmentClass.getConstructor(new Class[] {int.class});
		int surface = Integer.valueOf((String) config.getValue("surface")).intValue();
		NetworkEnvironment environment = (NetworkEnvironment) constructor.newInstance(new Object[] {new Integer(surface)});
		net.setNetworkEnvironment(environment);
		net.getNetworkEnvironment().configure();

		// creates the stations that will populate the network
		double density = Double.valueOf((String) config.getValue("density")).doubleValue();
		net.createStations(density);
		
		
		// the stations have been located randomly on the environment
		// it might be interested to let them run for a while in order
		// to stabilize the network graph
		net.getNetworkEnvironment().initializeNodesLocation(net.getAllComputers());

		// once we consider the location of the station has been initialized
		// we establish the connections that define the network itself
		net.createNewConnections();

		// the network will be handled by a simulation
		MadhocSimulation simulation = new MadhocSimulation();
		simulation.setName((String) config.getValue("simulation_name"));
		simulation.setResolution(1d / Double.valueOf((String) config.getValue("simulation_resolution")).doubleValue());
		simulation.setNetwork(net);




		// it is now time to define the applications that will run on the network
		{
			Collection applications = new Vector();

			{
				// the "core" application is a basic application that aim at extracting some
				// general purpose information on the simulation process
				Monitor app = new NetworkMonitor();
				applications.add(app);
				applications.add(new SimulationMonitor());
				app.configure();
			}
			{
				// the broadcasting application is what we want to run and observe
				Class protocol = Class.forName((String) config.getValue("broadcasting_protocol"));

				// define the broadcasting strategy we want to carry out
				BroadcastingMonitor broadcastingApp = new BroadcastingMonitor();
				broadcastingApp.setStationApplicationClass(protocol);
				broadcastingApp.setMode(BroadcastingMonitor.ONE_TO_ALL);
				applications.add(broadcastingApp);
				broadcastingApp.configure();
			}

			
			// the simulation knows the application it has to carry out
			// now let deploy them over the network
			simulation.deployApplications(applications);
			
			Iterator simulationAppIterator = simulation.getMonitorMap().values().iterator();
			
			while (simulationAppIterator.hasNext())
			{
				Monitor simulApp = (Monitor) simulationAppIterator.next();
				Iterator stationAppIterator = simulApp.getStationApplications().iterator();
				
				while (stationAppIterator.hasNext())
				{
					Application app = (Application) stationAppIterator.next();
					app.configure();
				}
			}
		}
		
		{
			Collection requiredMeasures = config.getValues("measures");
			
			if (requiredMeasures != null)
			{
				Iterator measureNames = requiredMeasures.iterator();
				Collection measures = new HashSet();
				
				while (measureNames.hasNext())
				{
					String name = (String) measureNames.next();
					Sensor m = simulation.findSensorByMeasureName(name);

					if (m == null)
						throw new IllegalArgumentException("measure '" + name + "' cannot be found");
					
					measures.add(m);
				}

				Iterator appIterator = simulation.getMonitorMap().values().iterator();
				
				while (appIterator.hasNext())
				{
					Monitor app = (Monitor) appIterator.next();
					app.getSensorMap().clear();
				}

				Iterator measureIterator = measures.iterator();
				
				while (measureIterator.hasNext())
				{
					Sensor m = (Sensor) measureIterator.next();
					m.getMonitor().addMeasure(m);
				}
			}
		}

		// at this step the simulation is ready to be used
		// the user can get information on the simulation process via the Simulation.getMeasures() method,
		// via the methods provided by the various objects (Simulation, Network, Station...)
		// or via the discrete event file, depending on his own needs
		BroadcastingMonitor broadcastingApp = (BroadcastingMonitor) simulation.getMonitorMap().get(BroadcastingMonitor.class);
		
		
		while (broadcastingApp.getAverageCoverageRate() < broadcastingApp.getTerminationCoverage())
		{
			if (simulation.getIteration() > 120 && ((NumericalSensor) broadcastingApp.getSensorMap().get(AverageNumberOfEmissionMeasure.class)).getVariationInTheLastIterations(120) == 0)
			{
				break;
			}

			simulation.iterate();
		}
		
		MeasureUtility.saveNumericalMeasuresToGNUPlotStream(simulation, config, System.out);
	}
}
