package org.lucci.madhoc.broadcast.impl.standard;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

import org.lucci.madhoc.broadcast.BroadcastInformation;
import org.lucci.madhoc.broadcast.LocalInfo;
import org.lucci.madhoc.broadcast.RandomDelayBasedBroadcastingProtocol;
import org.lucci.madhoc.messaging.Message;
import org.lucci.madhoc.messaging.TransferableObject;
import org.lucci.madhoc.network.Station;

/**
 * @author luc.hogie
 * Created on Jun 24, 2004
 */
public class MultiPointRelaying
	extends RandomDelayBasedBroadcastingProtocol
{
	public static class AHBPBroadcastInformation extends BroadcastInformation
	{
		// the ad needs to carry the list of MPRs
		public Collection mprs = new Vector();
		

		public Object clone()
		{
			AHBPBroadcastInformation clone = (AHBPBroadcastInformation) super.clone();
			clone.mprs.addAll(mprs);
			return clone;
		}

		
		public int getSizeInBytes()
		{
			return super.getSizeInBytes() + mprs.size() * 4;
		}
	}

	protected BroadcastInformation instantiateBroadcastInformation()
	{
		return new AHBPBroadcastInformation();
	}


	public String getName()
	{
		return "Multi-point Relaying";
	}

	/* (non-Javadoc)
	 * @see org.lucci.madhoc.broadcast.BroadcastingProtocol#isMultipleRebroadcastAllowed()
	 */
	public boolean isMultipleRebroadcastAllowed()
	{
		return false;
	}

	/* (non-Javadoc)
	 * @see org.lucci.madhoc.broadcast.RandomDelayBasedBroadcastingProtocol#takeDecision(org.lucci.madhoc.broadcast.BroadcastMessage)
	 */
	protected void takeDecision(TransferableObject object)
	{
		LocalInfo info = getLocalInfo(object);
		AHBPBroadcastInformation broadcastInfo = (AHBPBroadcastInformation) info.broadcastInformation;
	
		if (broadcastInfo.mprs.contains(getComputer()))
		{
			info.decision = 1;
		}
		else
		{
			info.decision = 0;
		}
	}

	
	
	
	protected void messageJustReceived(Message msg)
	{
		super.messageJustReceived(msg);
		takeDecision(msg.getContent());
	}

	/* (non-Javadoc)
	 * @see org.lucci.madhoc.broadcast.BroadcastingProtocol#adJustBeingSent(org.lucci.madhoc.broadcast.BroadcastMessage)
	 */
	protected void messageJustBeingSent(Message msg)
	{
		AHBPBroadcastInformation ahbpMsg = (AHBPBroadcastInformation) msg.getMessageInformation();
		ahbpMsg.mprs = getMPRSet();
	}
	
	
	protected Collection getMPRSet()
	{
		return getMPRGreedySet();
	}

	/**
	 * Starts with an empty set and tries to extend it with 1-hop neighbors
	 * that will improve the 2-hop coverage.
	 */
	protected Collection getMPRGreedySet()
	{
		Collection set = new Vector();
		
		while (true)
		{
			Station complement = getBestComplement(set);

			// no complement found, the greedy set is ready to be used
			if (complement == null)
			{
				return set;
			}
			else
			{
				set.add(complement);
			}
		}
	}
	
	
	/**
	 * Finds the best complement for the given set.
	 * The complement is a 1-hop neighbor that improves the 2-hop coverage.
	 */
	private Station getBestComplement(Collection currentSet)
	{
		Collection neighbors = getComputer().getNetworkingUnit().getNeighborhood();
		
		if (neighbors.isEmpty())
		{
			return null;
		}
		else
		{
			Station bestComplement = null;
			double originalPerf = getMPRSetPerformance(currentSet);
			double bestPerf = originalPerf;
			Iterator i = neighbors.iterator();
			
			while (i.hasNext())
			{
				Station testComplement = (Station) i.next();
				
				if (!currentSet.contains(testComplement))
				{
					Collection testSet = new HashSet(currentSet);
					testSet.add(testComplement);
					double perf = getMPRSetPerformance(testSet);
					
					if (perf > bestPerf)
					{
						bestPerf = perf;
						bestComplement = testComplement;
					}
				}
			}
			
			return bestComplement;
		}
	}

	private double getMPRSetPerformance(Collection mprSet)
	{
		Collection coverSet = getCoverSet(mprSet);
		return coverSet.size();
	}


	/**
	 * Returns the list of stations in the 1-hop neighborhood that can be reached if
	 * the given stations are used as relays.
	 * @param mprSet
	 * @return
	 */
	private Collection getCoverSet(Collection mprSet)
	{
		Collection coverSet = new HashSet();
		Iterator mprIterator = mprSet.iterator();
		
		while (mprIterator.hasNext())
		{
			Station relay = (Station) mprIterator.next();
			coverSet.add(relay);
			coverSet.addAll(relay.getNetworkingUnit().getNeighborhood());
		}

		return coverSet;
	}

}

