package org.lucci.madhoc.network.env.mall;

import java.awt.Color;
import java.util.Collection;
import java.util.HashSet;

import org.lucci.config.ConfigurationException;
import org.lucci.madhoc.env.NetworkEnvironment;
import org.lucci.math.Utilities;
import org.lucci.up.data.Figure;
import org.lucci.up.data.Point;
import org.lucci.up.data.rendering.point.CirclePointRenderer;
import org.lucci.up.system.Space;

/*
 * Created on Jul 29, 2004
 */

/**
 * @author luc.hogie
 */
public class HumanEnvironment extends NetworkEnvironment
{
	private Collection<Spot> spots = new HashSet<Spot>();

    
    public void configure()
		throws Throwable
	{
		super.configure();
        setRadioPropagationModel(new HumanRadioPropagationModel());
		getRadioPropagationModel().configure();
		
		double shopDensity = getNetwork().getSimulation().getConfiguration().getDouble("human_environment_spot_density");
        int shopCount = (int) Math.max(2, (shopDensity * getGrid().getEdgeLenght() * getGrid().getEdgeLenght()) / 1000000);
		double[] interval = getNetwork().getSimulation().getConfiguration().getInterval("human_environment_spot_radius");
		locateSpots(shopCount, interval[0], interval[1]);
    }

	public Collection<Spot> getSpots()
	{
		return spots;
	}


	protected void locateSpots(int shopCount, double minRadius, double maxRadius)
        throws ConfigurationException
	{
        String creationMode = getNetwork().getSimulation().getConfiguration().getString("human_mobility_spot_creation_method");
        
        if (creationMode.equalsIgnoreCase("random"))
        {
            double minimumDistanceBetweenSpots = getNetwork().getSimulation().getConfiguration().getDouble("human_mobility_minimum_distance_between_spots");
            double maximumDistanceBetweenSpots = getNetwork().getSimulation().getConfiguration().getDouble("human_mobility_maximum_distance_between_spots");

            while (shopCount-- > 0)
            {
                Spot newSpot = new Spot();
                newSpot.setRadius(Utilities.getRandomBetween(minRadius, maxRadius, getNetwork().getSimulation().getRandomNumberGenerator().getRandom()));
                Point testLocation = null;
                int attempt = (int) Math.pow(getGrid().getEdgeLenght(), 2); 

                while (testLocation == null)
                {
                    // locates the spot so as its bounds do not go of out the simulation bounds
                    double x = Utilities.getRandomBetween(newSpot.getRadius(), getGrid().getEdgeLenght() - newSpot.getRadius(), getNetwork().getSimulation().getRandomNumberGenerator().getRandom());
                    double y = Utilities.getRandomBetween(newSpot.getRadius(), getGrid().getEdgeLenght() - newSpot.getRadius(), getNetwork().getSimulation().getRandomNumberGenerator().getRandom());
                    testLocation = new Point(x, y);
                    
                    // then checks the distance to all other spots
                    for (Spot otherSpot : getSpots())
                    {
                        double distance = otherSpot.getLocation().getDistanceTo(testLocation) - newSpot.getRadius() - otherSpot.getRadius();

                        // if the new spot is too close or too far
                        if (distance < minimumDistanceBetweenSpots || distance > maximumDistanceBetweenSpots)
                        {
                            // test another position
                            testLocation = null;
                            break;
                        }
                    }

                    if (--attempt == 0)
                    {
                        throw new IllegalStateException("the configuration you provided for the human environment does not permit the simulator to initialize it");
                    }
                }

                newSpot.setLocation(testLocation);
                getSpots().add(newSpot);
            }
        }
        else if (creationMode.equalsIgnoreCase("grid"))
        {
            int rowCount = (int) Math.sqrt(shopCount);
            double distanceBetweenTwoSpots = getGrid().getEdgeLenght() / rowCount;
            double alteration = getNetwork().getSimulation().getConfiguration().getDouble("human_mobility_grid_alteration_ratio") * getNetwork().getNetworkEnvironment().getGrid().getEdgeLenght();
            
            for (int i = 0; i < rowCount; ++i)
            {
                for (int j = 0; j < rowCount; ++j)
                {
                    Spot newSpot = new Spot();
                    newSpot.setRadius(Utilities.getRandomBetween(minRadius, maxRadius, getNetwork().getSimulation().getRandomNumberGenerator().getRandom()));
                    double x = i * getGrid().getEdgeLenght() / rowCount + distanceBetweenTwoSpots / 2;
                    double y = j * getGrid().getEdgeLenght() / rowCount + distanceBetweenTwoSpots / 2;
                    x += Utilities.getRandomBetween(-alteration, alteration, getNetwork().getSimulation().getRandomNumberGenerator().getRandom());
                    y += Utilities.getRandomBetween(-alteration, alteration, getNetwork().getSimulation().getRandomNumberGenerator().getRandom());
                    newSpot.setLocation(new Point(x, y));
                    getSpots().add(newSpot);
                }
            }
        }
        else
        {
            throw new ConfigurationException("human_mobility_spot_creation_method", "unknown creation mode: " + creationMode);
        }
	}
    
//	private double getMinimalDistanceBetweenSpots()
//	{
//		double minD = Integer.MAX_VALUE;
//		Iterator i1 = getSpots().iterator();
//		
//		while (i1.hasNext())
//		{
//			Spot s1 = (Spot) i1.next();
//
//			Iterator i2 = getSpots().iterator();
//		
//			while (i2.hasNext())
//			{
//				Spot s2 = (Spot) i2.next();
//				
//				if (s1 != s2)
//				{
//					double d = s1.getLocation().getDistanceTo(s2.getLocation());
//					
//					if (d < minD )
//					{
//						minD = d;
//					}
//				}
//			}
//		}
//
//		return minD;
//	}


	
	public String toString()
	{
		return "human-environment";
	}

	/* (non-Javadoc)
	 * @see org.lucci.madhoc.env.NetworkEnvironment#createFigure(org.lucci.up.system.Space)
	 */
	public Figure createFigure(Space space)
	{
		Figure f = new Figure();
		
		for (Spot spot : getSpots())
		{
			Point shopPoint = new Point(spot.getLocation().getX(), spot.getLocation().getY());
			CirclePointRenderer r = new CirclePointRenderer();
			r.setColor(Color.red);
			int pixelRadius = (int) (spot.getRadius() / space.getXDimension().getLogicalIntervalRepresentedByOnePixel());
			r.setRadius(pixelRadius);
			shopPoint.addRenderer(r);
			f.addPoint(shopPoint);
		}
		
		return f;
	}

    @Override
    public Class getDefaultMobilityModel()
    {
        return HumanMobility.class;
    }

}
